
/// Includes
#include "bkapi.h"
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/ports.h>
#include <dos/dos.h>
#include <mmu/mmubase.h>
#include <mmu/context.h>
#include <mmu/mmutags.h>
#include <proto/exec.h>
#include <proto/mmu.h>
#include <proto/dos.h>
#include <string.h>
#include <stdio.h>
///
/// Defines
int __nocommandline=1;
extern struct MMUBase *MMUBase;
struct bkerhss *glob_ss = NULL;
#ifdef BKDEBUG
#  define D(a) (a);
#else
#  define D(a) ;
#endif
///
/// Statics
struct MMUBase *MMUBase;
struct DosLibrary *DOSBase;
struct ExecBase *SysBase;
///
/// Protos
ULONG er_init(void);
ULONG ASM er_availmem(REG(d1,ULONG));
ULONG ASM er_getarea(REG(a0,ULONG *));
int ProtectMem(struct MMUContext *,struct MMUContext *,ULONG,ULONG);
///
/// er_init
/****** bkapi.o/er_init ******************************************************
*
*   NAME
*   er_init -- initialize EXTRES buffer use
*
*   SYNOPSIS
*   version = er_init()
*   D0
*
*   ULONG er_init();
*
*   FUNCTION
*   Initializes use of EXTRES buffer memory pool of BlizKick. This
*   routine must be called before using any of the other functions.
*
*   RESULT
*   version - version number of EXTRES buf API available or zero
*       if EXTRES buf could not be found (ie. BlizKick not run).
*       Currently ERH_API_V1.
*
*   SEE ALSO
*
******************************************************************************
*
*/

ULONG er_init(void) {
  struct bkerhss *ss;

  if (!glob_ss) {

    Forbid();
    ss = (struct bkerhss *) FindSemaphore("EXTRES Handler");
    Permit();

    if (ss) {

      switch (ss->version) {

        case ERH_API_V1:
          glob_ss = ss;
          break;

        /*
        case ERH_API_V2:
          glob_ss = ss;
          ... could do some version dependant check / init here ...
          break;
        */
      }

      /* force sanity check */
      er_availmem(MEMF_LARGEST);

    }
  }
  return glob_ss ? glob_ss->version : 0;
}
///
/// er_availmem
/****** bkapi.o/er_availmem **************************************************
*
*   NAME
*   er_availmem -- return EXTRES buffer memory available
*
*   SYNOPSIS
*   size = er_availmem(requirements)
*   D0         D1
*
*   ULONG er_availmem(ULONG);
*
*   FUNCTION
*   This function returns the amount of free EXTRES buffer memory.
*
*   To find out what the largest block is, specify MEMF_LARGEST in
*   attributes argument.
*
*   WARNING
*   Due to the effect of multitasking, the value returned may not
*   actually be the amount of free memory available at that instant.
*   However if you have locked memory with er_lock() before, the
*   result is exact.
*
*   INPUTS
*   requirements - MEMF_LARGEST results calculation of the size of
*       the largest block.
*
*   RESULT
*   size - total free space remaining (or the largest free block).
*
*   NOTE
*   er_availmem(MEMF_LARGEST) does a consistency check on the
*   memory list. In case of an error 0xFFFFFFFF is returned, and
*   further allocations are impossible. If memory header itself is
*   bad will alert with BKA_MemoryInsane.
*
*   SEE ALSO
*   exec.library/AvailMem, exec/memory.h
*
******************************************************************************
*
*/

ULONG ASM er_availmem(REG(d1,ULONG attributes)) {
  struct MemHeader *mh;
  struct MemChunk *mc, *predmc;
  ULONG ret = NULL, free = 0, failed = 0;

  if (glob_ss) {
    ObtainSemaphore(&glob_ss->ss);

    if (attributes & MEMF_LARGEST) {

      mh = glob_ss->mh;

      if ( (mh->mh_Node.ln_Succ == NULL) &&
           (mh->mh_Node.ln_Pred == NULL) &&
           (mh->mh_Node.ln_Type == NT_MEMORY) &&
           (((ULONG) mh->mh_Lower & MEM_BLOCKMASK) == 0) &&
           (((ULONG) mh->mh_Upper & MEM_BLOCKMASK) == 0) &&
           (mh->mh_Lower < mh->mh_Upper)
         ) {

        mc = mh->mh_First;
        predmc = (struct MemChunk *) ((ULONG) mc - 1);

        while (mc && (!failed) ) {
          if ( ( ((ULONG) mc & MEM_BLOCKMASK) == 0) &&
               ((ULONG) mc > (ULONG) predmc) &&
               ((ULONG) mc <= ((ULONG) mh->mh_Upper - MEM_BLOCKSIZE))
             ) {

            if ( (mc->mc_Bytes & MEM_BLOCKMASK) == 0) {

              free += mc->mc_Bytes;
              if (mc->mc_Bytes > ret) ret = mc->mc_Bytes;

            } else failed = 1;

            predmc = mc;
            mc = mc->mc_Next;

          } else failed = 1;

        }
      } else {
        failed = 2;
      }

      if ( (!failed) && (free != mh->mh_Free) ) failed = 1;

      switch (failed) {
        case 2:
          /* memheader screwed up!! */
          D(Printf("er_availmem(largest): *** memheader fucked ***!!\n"));
          Alert(BKA_MemoryInsane);
          /* fall thru */
        case 1:
          D(Printf("er_availmem(largest): freelist fucked!\n"));
          mh->mh_First = NULL;
          mh->mh_Free = 0;
          ret = 0xFFFFFFFF;
          break;
      }

    } else {
      ret = glob_ss->mh->mh_Free;
    }

    ReleaseSemaphore(&glob_ss->ss);
  }
  return ret;
}
///
/// er_getarea
/****** bkapi.o/er_getarea ***************************************************
*
*   NAME
*   er_getarea -- get EXTRES resident module area
*
*   SYNOPSIS
*   start = er_getarea(len_ptr)
*   D0         A0
*
*   void *er_getarea(ULONG *);
*
*   FUNCTION
*   Return lower and upper bound of memory area covered by EXTRES
*   buffer.
*
*   INPUTS
*   len_ptr - pointer to ULONG to put area lenght to.
*
*   RESULT
*   start - pointer to start of resident module area. If there's no
*       BlizKick EXTRES buffer available will be zero.
*
*   NOTE
*   You *must* er_lock() before you read/write EXTRES buffer memory
*   area. Call er_unlock() when done tempering with it.
*
*   Will return zero if BlizKick is not run.
*
*   SEE ALSO
*   er_lock, er_unlock
*
******************************************************************************
*
*/

ULONG ASM er_getarea(REG(a0,ULONG *len_ptr)) {
  ULONG ret = NULL;

  if (glob_ss) {
    ObtainSemaphore(&glob_ss->ss);

    ret = (ULONG) glob_ss->mh->mh_Lower;
    *len_ptr = (ULONG) glob_ss->mh->mh_Upper - ret;

    ReleaseSemaphore(&glob_ss->ss);
  }
  return ret;
}
///
/// main
int main(void) {
    ULONG psize,ver,size,start,end;
    struct MMUContext *ctx,*sctx;

    printf("MuProtectERH $VER: 1.0 (29/07/2001) by Sante Nocciolino\n\n");

    SysBase=*((struct ExecBase **)(4L));
    if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37L))) {
        if ((MMUBase=(struct MMUBase *)OpenLibrary("mmu.library",41L))) {

    ctx=DefaultContext();   /* get the default context */
    sctx=SuperContext(ctx); /* get the supervisor context for this one */

    psize=GetPageSize(ctx);
    if (psize!=GetPageSize(sctx)) {
                printf("Different user and supervisor page sizes are not suported.\n");
                return ERROR_OBJECT_WRONG_TYPE;
    }

    ver=er_init();

    if (ver) {
        start=er_getarea(&size);
        end=start+size-1;
        printf("Blizkick EXTRES Handler found!\n\n");
        printf("- start address\t$%08lx\n- end address\t$%08lx\n- buffer size\t$%08lx\n",start,end,size);
        printf("- page size\t$%08lx\n\n",psize);
        if (start & (psize-1)) {
            printf("Failed: start address is not aligned to MMU page size...\n");
            printf("To align start address to MMU page size you should add %ld bytes\nto the EXTRESBUF buffer in the BlizKick command line.\n",(start & (psize -1)));
            return(10);
        } else printf("Start address is aligned to MMU page size...\n");

        if ((end+1) & (psize-1)) {
            printf("Failed: end address is not aligned to MMU page size...\n");
            return(10);
        } else printf("End address is aligned to MMU page size...\n");

        if (ProtectMem(ctx,sctx,start,end)) {
            printf("Failed: buffer not protected.\n");
            return(10);
        } else {
            printf("Buffer successful protected.\n");
            return(0);
        }
    }
    printf("Failed: Blizkick EXTRES Handler not found!\n");
    return(10);
    }
    printf("Failed: I need mmu.library V41+.\n");
    return(10);
    }
    printf("Failed: I need dos.library V37+.\n");
    return(10);
}
///
/// ProtectMem
int ProtectMem(struct MMUContext *ctx,struct MMUContext *sctx,ULONG start,ULONG end) {

    struct MinList *ctxl,*sctxl;
    int err = ERROR_NO_FREE_STORE;

        LockContextList();
        LockMMUContext(ctx);
        LockMMUContext(sctx);

         if ((ctxl=GetMapping(ctx))) {
         if ((sctxl=GetMapping(sctx))) {

                if (SetProperties(ctx,MAPP_ROM,MAPP_WRITEPROTECTED|MAPP_ROM|MAPP_REPAIRABLE,start,end-start+1,0L)) {
                    if (SetProperties(sctx,MAPP_ROM,MAPP_WRITEPROTECTED|MAPP_ROM|MAPP_REPAIRABLE,start,end-start+1,0L)) {
                        if (RebuildTree(ctx) && RebuildTree(sctx)) err = 0;
                    }
                }

                if (err) {
                        SetPropertyList(ctx,ctxl);
                        SetPropertyList(sctx,sctxl);
                }

                ReleaseMapping(sctx,sctxl);
          }
          ReleaseMapping(ctx,ctxl);
         }

        UnlockMMUContext(sctx);
        UnlockMMUContext(ctx);
        UnlockContextList();

        return err;
}
///

